# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import inspect
import requests
import json
import copy
from pprint import pprint
from vcsconn.graphql_query import GraphQL_Query
from vcsconn.utils.eif_exception import OperationalError, EIFException
from vcsconn.utils.klog          import getCaller

__version__ = '2.0.5'

REPOS_PAGE_SIZE         = 10
PULL_REQUESTS_PAGE_SIZE = 30
COMMIT_PAGE_SIZE        = 30

class GitHubGraphQLAdapter:
    def __init__(self, config, logger):
        self.logger     = logger
        self.protocol   = config.get('Protocol', 'https')
        self.server     = config.get('Server',   'api.github.com')
        self.enterprise = False
        if self.server !=  'api.github.com':
            self.enterprise = True
        self.port       = config.get('Port',     443)
        self.user       = config.get('User',     None)
        self.password   = config.get('Password', None)
        self.token      = config.get('Token',    None)
        self.organization = config.get('Organization', None)
        #self.pagesize  = config.get('Pagesize', 100)
        #self.max_commits = config.get('MaxCommits', 200)

        # if enterprise, change https://github-isl-test-01.ca.com/api:443/graphql to https://github-isl-test-01.ca.com:443/api/graphql
        self.svc_name  = 'graphql'
        if self.enterprise:
            self.svc_name = 'api/graphql'
        self.query_url = "{0}://{1}:{2}/{3}".format(self.protocol, self.server, self.port, self.svc_name)
        self.connect()

    def connect(self):
        if self.user and self.token:
            self.auth = (self.user, self.token)
        elif self.user and self.password:
            self.auth = (self.user, self.password)
        else:
            return False

    def inhaleCommitNodesInfo(self, nodes, target_repos, ref_time, pail):
        repo_keychain = ['data', 'search']
        active_repos = [item for item in nodes if item['name'] in target_repos]
        # discount any repo block where repo_name not in target_repositories
        # if a repo block has a commits block with a pageInfo['nextPage'] : True, get pageInfo['endCursor'] to use as value for 'after:' property
        # pull out the commits on this page for a target_repository
        # call out to method that will get all commit pages , we already have page #1 here in the page we just retrieved
        pages = 0
        for repo_data in active_repos:
            repo_name = repo_data['name']
            # pull the first page of commit info out, note if commits hasNextPage is True
            repo_commits, next_page, commits_cursor = self.extractCommitInfo(repo_data['ref']['target']['history'])
            if not repo_commits:
                continue
            pail[repo_name] = [commit['node'] for commit in repo_commits]
            pages += 1

            while next_page:
                page_commits, next_page, commits_cursor = self.getAdditionalCommitPage(repo_name, commits_cursor,
                                                                                       repo_keychain[1:], ref_time)
                pages += 1
                if page_commits:
                    pail[repo_name].extend(page_commits)

        self.logger.debug("number of pages required to obtain commit info: %d" % pages)


    def getRepoAndCommitPages(self, target_repos, ref_time, branch='master'):
        active_repos_query  = GraphQL_Query.constructRepoActivityQuery(self.organization, ref_time, REPOS_PAGE_SIZE, COMMIT_PAGE_SIZE, branch='master')

        commit_keychain  = ['ref','target','history']
        pail   = {}
        first_page_query = active_repos_query.replace('<<nextPageAfterCursor>>', '')
        try:
            result = self.makeQueryRequest(first_page_query)
            nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
            self.inhaleCommitNodesInfo(nodes, target_repos, ref_time, pail)
        except Exception as ex:
            ex.report()
            raise

        while pageInfo['hasNextPage']:
            self.logger.debug("Last Cursor %s" % last_cursor)
            cursor_spec = ', after:"%s"' % last_cursor
            next_page_query = active_repos_query.replace('<<nextPageAfterCursor>>', cursor_spec)
            try:
                result = self.makeQueryRequest(next_page_query)
                nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
                self.inhaleCommitNodesInfo(nodes, target_repos, ref_time, pail)
            except Exception as ex:
                ex.report()
                raise

        return pail


    def getAdditionalCommitPage(self, repo, end_cursor, truncated_repo_chain, ref_time):
        query = GraphQL_Query.constructCommitsQuery(self.organization, repo, ref_time, COMMIT_PAGE_SIZE, branch='master', cursor=end_cursor)
        try:
            result = self.makeQueryRequest(query)
            commit_page = result['data']['repository']['ref']['target']['history']
            nodes, pageInfo, end_cursor = self._consumeAPage(commit_page)
        except Exception as ex:
            ex.report()
            raise
        # on all pages except first we need to restore dict structure as below (just so that extractCommitInfo works for all pages)
        return self.extractCommitInfo({'edges' : nodes, 'pageInfo' : pageInfo})


    def makeQueryRequest(self, graphql_query):
        stack_item = inspect.stack()[1]  # assume caller of interest is stack_levels down in the call stack
        frame, filename, line_no, code_unit = stack_item[:4]
        caller = getCaller(frame)
        self.logger.debug("GitHub GraphQL query on behalf of %s:\n%s" % (caller, graphql_query))
        response = requests.post(self.query_url, json.dumps({"query": graphql_query}), auth=self.auth)
        #response = requests.post(self.query_url, json.dumps({"query": graphql_query}), auth=self.auth, verify=False)
        if response.status_code >= 400:
            raise OperationalError('GraphQL request resulted in an error')
        result = response.json()

        if 'errors' in result:
            msg = result['errors'][0]['message']
            raise OperationalError(msg)
        return result

    def getAllQueryPages(self, query):
        bucket = []
        first_page_query = query.replace('<<nextPageAfterCursor>>', '')
        pages = 0
        try:
            result = self.makeQueryRequest(first_page_query)
            nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
            bucket.extend(nodes)
            pages += 1
        except EIFException as ex:
            ex.report()
            raise
        except Exception as ex:
            raise


        while pageInfo['hasNextPage']:
            #self.logger.debug("Last Cursor %s" % last_cursor)
            cursor_spec = ', after:"%s"' % last_cursor
            next_page_query = query.replace('<<nextPageAfterCursor>>', cursor_spec)
            try:
                result = self.makeQueryRequest(next_page_query)
                nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
                bucket.extend(nodes)
                pages += 1
            except Exception as ex:
                ex.report()
                raise

        self.logger.debug("number of pages required to satisfy the query: %d" % pages)
        return bucket

    def extractCommitInfo(self, commit_data):
        commits    = commit_data['edges']
        page_info  = commit_data['pageInfo']
        next_page  = page_info['hasNextPage']
        end_cursor = page_info['endCursor']
        return commits, next_page, end_cursor


    def extractPullRequestInfo(self, pr_data):
        pull_requests = pr_data['edges']
        return pull_requests

    def _consumeAPage(self, chunk):
        items = chunk['edges']
        nodes = [i['node'] for i in items]
        pageInfo  = chunk['pageInfo']
        if items:
            end_cursor   = pageInfo['endCursor']
        else:
            end_cursor = None
        return nodes, pageInfo, end_cursor

    def getPullRequestPages(self, target_repos, ref_time, branch='master'):
        active_repos_query  = GraphQL_Query.constructPullRequestsQuery(self.organization, ref_time, REPOS_PAGE_SIZE, PULL_REQUESTS_PAGE_SIZE)
        repo_keychain    = ['data', 'search']
        #commit_keychain  = ['ref','target','history']
        pr_keychain = ['target', 'repository', 'pullRequests']
        pail   = {}
        first_page_query = active_repos_query.replace('<<nextPageAfterCursor>>', '')
        pages = 0

        try:
            result = self.makeQueryRequest(first_page_query)
            nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
            self.inhalePullRequestNodesInfo(nodes, target_repos, pail)
            pages += 1

        except Exception as ex:
            ex.report()
            raise

        while pageInfo['hasNextPage']:
            #self.logger.debug("Last Cursor %s" % last_cursor)
            cursor_spec = ', after:"%s"' % last_cursor
            next_page_query = active_repos_query.replace('<<nextPageAfterCursor>>', cursor_spec)
            try:
                result = self.makeQueryRequest(next_page_query)
                nodes, pageInfo, last_cursor = self._consumeAPage(result['data']['search'])
                self.inhalePullRequestNodesInfo(nodes, target_repos, pail)
                pages += 1
            except Exception as ex:
                ex.report()
                raise
        self.logger.debug("number of pages required to obtain the pull requests: %d" % pages)
        return pail

    def inhalePullRequestNodesInfo(self, nodes, target_repos, pail):
        # discount any repo block where repo_name not in target_repositories
        # pull out the pull request on first page for a target_repository
        active_repos = [item for item in nodes if item['name'] in target_repos]
        for repo_data in active_repos:
            repo_name = repo_data['name']
            #pull_requests = self.extractPullRequestInfo(repo_data['target']['repository']['pullRequests'])
            pull_requests = self.extractPullRequestInfo(repo_data['ref']['target']['repository']['pullRequests'])
            pail[repo_name] = [pr['node'] for pr in pull_requests]
