# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import textwrap

class GraphQL_Query:

    @staticmethod
    def constructOrganizationsQuery():
        query = \
        """
        query {
          viewer {
            login
            organizations(first: 10) { edges { org:node { name resourcePath } } }
          }
        }
        """
        return textwrap.dedent(query[1:])

    @staticmethod
    def constructOrganizationRepositoriesQuery(organization):
        query = \
            """
              {search(first: 100, type: REPOSITORY, query: "user:%s"<<nextPageAfterCursor>>) {
                edges {
                  node {
                    ... on Repository { id name }
                  }
                  cursor
                }
                pageInfo{endCursor hasNextPage}
                repositoryCount
              }
            }
            """ % (organization)
        return textwrap.dedent(query[1:])


    @staticmethod
    def constructRepoActivityQuery(organization, ref_time, pagesize, commit_pagesize, branch='master', cursor=None):
        """
            construct a GraphQL compliant query to specify commits in repos that happened since a specific time.
        """
        example_query_not_really_used = \
            """
              {search(first: %s, type: REPOSITORY, query: "user:%s pushed:>%s"<<nextPageAfterCursor>>) {
                edges {
                  node {
                    ... on Repository {
                      id
                      name
                      pushedAt
                      ref(qualifiedName:"%s") {
                        target {
                          ... on Commit {
                            history(first: %d, since:"%s"){
                              edges {
                                node { repository{name} oid message committer{name email} committedDate resourcePath}
                                cursor
                              }
                            }
                          }
                        }
                      }
                     }}}
                     cursor
                  }
                  pageInfo{endCursor hasNextPage}
                  repositoryCount
                }
            }
            """ % (pagesize, organization, ref_time.asISOString(), branch, commit_pagesize, ref_time.asISOString())

        commit_spec     = GraphQL_Query._commitSpec(ref_time, commit_pagesize)[9:]
        repository_spec = GraphQL_Query._repositorySpec(branch)[8:]
        search_spec     = GraphQL_Query._repoSearchSpec(organization, ref_time, branch, 100)
        full_repo_spec  = repository_spec.replace('<< commit_spec >>',        commit_spec)
        full_query      = search_spec.replace('<< repository_spec >>', full_repo_spec)
        return "{%s\n}" % textwrap.indent(full_query, '  ')

    @staticmethod
    def _repoSearchSpec(owner, ref_time, branch, pagesize, cursor=None):
        search_spec = \
        """
        search(first: %d, %s type: REPOSITORY, query: "user:%s pushed:>%s"<<nextPageAfterCursor>>)
        {
          edges
          {
            node
            {
               << repository_spec >>
            }
            cursor
          }
          pageInfo {endCursor hasNextPage}
          repositoryCount
        }"""
        cursor_clause = "after: %s, " if cursor else ''
        result = search_spec % (pagesize, cursor_clause, owner, ref_time.asISOString())
        return textwrap.dedent(result)

    @staticmethod
    def _repositorySpec(branch):
        repository_spec = \
        """
        ... on Repository
        {
          id
          name
          pushedAt
          ref(qualifiedName:"%s")
          {
            << commit_spec >>
          }
        }"""
        return repository_spec % branch

    @staticmethod
    def _commitSpec(ref_time, pagesize):
        commit_spec = \
        """
        target
            {
              ... on Commit
              {
                history(first: %d, since:"%s")
                {
                  edges
                  {
                    node  { oid repository{name} message committer {name email} committedDate resourcePath }
                    cursor
                  }
                  pageInfo {endCursor hasNextPage}
                }
              }
            }"""

        return commit_spec % (pagesize, ref_time.asISOString())


    @staticmethod
    def constructCommitsQuery(organization, repo, ref_time, pagesize, branch='master', cursor=None):
        """
            construct a GraphSQL compliant query to get commit related info for commits
            that were done past a specific time on a specific branch.
        """
        cursor_clause = ', after: "%s"' % cursor if cursor else ''
        query = \
            """
            { repository(owner: "%s", name: "%s") {
                ... on Repository {
                  ref(qualifiedName:"%s") {
                    target {
                      ... on Commit {
                        history(first: %s, since:"%s"%s) {
                          edges {
                            node { oid repository{name} message committer {name email} committedDate resourcePath }
                            cursor
                          }
                          pageInfo {endCursor hasNextPage}
                        }
                      }
                    }
                  }
                }
              }
            }
            """  % (organization, repo, branch, pagesize, ref_time.asISOString(), cursor_clause)
        return textwrap.dedent(query[1:])

    # @staticmethod
    # def _pullRequestSpec(ref_time):
    #     pr_spec = \
    #     """
    #       pullRequests(first:10){
    #           edges{
    #             node{
    #               title
    #               number
    #               comments(last:1){
    #                 edges{
    #                   node{
    #                     body
    #                   }
    #                 }
    #               }
    #               commits(first:100){
    #                 nodes{
    #                   url
    #                   commit{message}
    #                 }
    #               }
    #             }
    #           }
    #         }"""
    #
    #     return pr_spec % (ref_time)

    @staticmethod
    #def constructPullRequestsQuery(organization, ref_time, pagesize, cursor=None):
    def constructPullRequestsQuery(organization, ref_time, repos_pagesize, pull_requests_pagesize, cursor=None):
        """
        construct a GraphQL compliant query to specify pull requests in repos that happened since a specific time.
        """
        pull_requests_full_query = \
        """
            { search(first: %s, type: REPOSITORY, query: "user:%s pushed:>%s"<<nextPageAfterCursor>>) {
                edges {
                  node {
                    ... on Repository {
                      name
                      pushedAt
                      ref(qualifiedName:"master"){
                        target{
                          repository{
                            name
                            pullRequests(first:%s orderBy: {field: CREATED_AT, direction: DESC}){
                              edges{
                                node{
                                  title
                                  url
                                  createdAt
                                  number
                                  state
                                  comments(first:1){
                                    edges{
                                      node{
                                        body
                                      }
                                    }
                                  }
                                  commits(first:10){
                                    nodes{
                                      commit{oid message}
                                    }
                                  }
                                }
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                  cursor
                }
                pageInfo{endCursor hasNextPage}
                repositoryCount
              }
            }
            """ % (repos_pagesize, organization, ref_time.asISOString(), pull_requests_pagesize )

        #pr_spec = GraphQL_Query._pullRequestSpec(ref_time)[9:]
        #repository_spec = GraphQL_Query._repositorySpec(branch)[8:]
        #search_spec = GraphQL_Query._repoSearchSpec(organization, ref_time, branch, 100)
        #full_repo_spec = repository_spec.replace('<< pr_spec >>', pr_spec)
        #full_query = search_spec.replace('<< repository_spec >>', full_repo_spec)
        #return "{%s\n}" % textwrap.indent(pull_requests_full_query, '  ')
        return pull_requests_full_query