# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import requests

from vcsconn.utils.eif_exception import ConfigurationError, OperationalError
from vcsconn.tfsgit_changeset    import TFSGitChangeset
from vcsconn.tfsgit_pullrequest  import TFSGitPullRequest
from vcsconn.utils.chronuti      import TimeStamp

class TFSGitHelper:

    STD_PAGESIZE   =  100
    REPOS_PAGESIZE = 1000

    def __init__(self, username, token, logger, proxy):
        self.username = username
        self.token    = token
        self.log      = logger
        self.http_proxy = proxy

        if self.username and self.token:
            self.auth = (self.username, self.token)

    def getFromTFSGit(self, uri):  # internal helper method that makes requests to TFSGit Server REST API
        result = None
        try:
            headers = {'Content-Type': 'application/json'}
            self.log.debug("uri: %s, proxy: %s" % (uri, self.http_proxy) )

            response = requests.get(uri, auth=self.auth, proxies=self.http_proxy, headers=headers)
            self.log.debug("TFSGit response code: '%s'" % response.status_code)
            if response.status_code > 299:
                msg = "Problem communicating with TFSGit: %s" % response.text
                raise OperationalError(msg)
            result = response.json()
        except Exception as exc:
            actual_error = exc.args[0]
            problem = "Unable to retrieve information from TFSGit. %s" % actual_error
            raise OperationalError(problem)
        return result

    def getAllRepos(self, repos_url):
        try:
            tfsgit_repos = self.getFromTFSGit(repos_url)
        except Exception as e:
            self.log.error(e)
            problem = "Unable to interrogate TFSGit for information on repositories: %s" % (e)
            raise OperationalError(problem)
        return tfsgit_repos['value']

    def getAllBranches(self, baseurl):
        # Get info about all the branches in the repo

        branches_url = "%s/refs/heads" % baseurl
        self.log.debug('Branch URL: ' + branches_url)
        branches = self.getFromTFSGit(branches_url)

        branch_data = [{item['name'] : item['objectId']} for item in branches['value']]
        return branch_data

    def getQualifiedCommits(self, url, repo):
        commits = []
        try:
            results = self.getFromTFSGit(url)
        except Exception as ex:
            ex.report()
            return ([], '')

        for commit in results['value']:
            commit['repository'] = repo
            commits.append(TFSGitChangeset(commit))

        return commits

    def getQualifiedPullRequests(self, pr_url, ref_time):
        pull_requests = []
        try:
            results = self.getFromTFSGit(pr_url)
        except Exception as ex:
            ex.report()
            return []
        for pr in results['value']:
            if TimeStamp.fromFormattedString(pr['creationDate'][:-5] + 'Z') > ref_time:
                pull_request = TFSGitPullRequest(pr)
                pull_request.tfsgit_helper = self
                pull_requests.append(pull_request)
        return pull_requests
