# Copyright 2017, 2018 CA Technologies.  All Rights Reserved.

#################################################################################################
#
# auxload  - Auxiliary functionality loader
#              Contains classes to load auxiliary (plugin and extension) classes
#
#################################################################################################

import sys, os
import re
import importlib


#################################################################################################

class AuxLoader(object):
    def __init__(self):
        #self.facility_class_name = ""
        #self.facility_class = None
        self.facility = {}  # key by facility_class_name, value is facility_class object

    def loadFacility(self, facility_type, facility_class_name, subdir=None, module_name=None):
        """
            A specification looks like:
                                                   method_name
                            module_name.class_name.method_name
                  directory.module_name.class_name.method_name
        """
        conventional_facility_dir = facility_type.lower()
        subdir = subdir if subdir else conventional_facility_dir

        if not os.path.exists(subdir):
            raise Exception('No such %s subdir: %s found' % (facility_type, subdir))
        if not os.path.isdir(subdir):
            raise Exception('No such %s subdir: %s found' % (facility_type, subdir))

        if module_name:
            mod_path = '%s/%s.py' % (subdir, module_name)
            if not os.path.isfile(mod_path):
                raise Exception('No such %s module: %s found' % (facility_type, mod_path))

        def qualified(subdir, fn):
            return fn.endswith('.py') and os.path.isfile('%s/%s' % (subdir, fn))

        pyfiles = [fn for fn in os.listdir(subdir) if qualified(subdir, fn)]
        module_names = [os.path.splitext(fn)[0] for fn in pyfiles]
        module_names = [os.path.splitext(module_name)[0]] if module_name else module_names

        target_classes_found = 0
        tc_module = {}  # target class module lookup dict
        # look in all modules, record the module name if the facility_class_name is defined in there
        for module_name in module_names:
            module_file = '%s/%s.py' % (subdir, module_name)

            with open(module_file, 'r') as mf:
                target_class_def = [line for line in mf if re.search('^class %s\s*\(' % facility_class_name, line)]

                if target_class_def:
                    tc_module[facility_class_name] = module_name
                    target_classes_found += 1
        if not target_classes_found:
            problem = "No class: '%s' defined in any module located in the '%s' subdir"
            raise Exception(problem % (facility_class_name, subdir))
        if target_classes_found > 1:
            problem = "%s class: '%s' defined in more than one module located in the '%s' subdir"
            raise Exception(problem % (facility_type, facility_class_name, subdir))

        facility_dir = subdir  # just in case later we widen the possibilities...
        sys.path.insert(0, facility_dir)
        # print "\n".join(sys.path)
        #module = __import__(tc_module[facility_class_name], globals(), None, [facility_class_name], 0)
        module = importlib.import_module(tc_module[facility_class_name]) #, facility_class_name
        del sys.path[0]

        facility_class = getattr(module, facility_class_name)

        facility_method_name = 'service'  # by convention

        if getattr(facility_class, facility_method_name):
            self.facility[facility_class_name] = facility_class
        else:
            print("unable to validate that the %s class has a %s method" % \
                  (facility_class_name, facility_method_name))

    def getFacility(self, facility_type, facility_class_name, subdir=None, module_name=None):
        """
            Given a facility_class_name, look into self.facility for a key match.
            If found return the corresponding value, which must be the
            class object for the facility_class.
        """
        if facility_class_name not in self.facility:
            self.loadFacility(facility_type, facility_class_name, subdir, module_name=module_name)

        return self.facility.get(facility_class_name, None)


#################################################################################################

class PluginLoader(AuxLoader):
    def loadPlugin(self, plugin_class_name):
        """
            A plugin_class_name looks like FoobarDoesDeltas and by convention should be
            defined in a module within the 'plugin' subdirectory.
        """
        super(PluginLoader, self).loadFacility('Plugin', plugin_class_name)

    def getPlugin(self, plugin_class_name):
        """
            Given a plugin_class_name as a String, return the corresponding class
            if such a class exists in the conventional location.
        """
        return super(PluginLoader, self).getFacility('Plugin', plugin_class_name)


##########################################################################################################

class ExtensionLoader(AuxLoader):
    def loadExtension(self, extension_spec):
        """
            An extension_class_name looks like FoobarDoesDeltas and by convention should be
            defined in a module within the 'extension' subdirectory, but can exist in the
            current working directory.
        """
        super(ExtensionLoader, self).loadFacility('Extension', extension_class_name)

    def getExtension(self, extension_spec):
        """
            Given an extension_class_name as a String, return the corresponding class
            if such a class exists in the conventional location.
        """
        ##        print "ExtensionLoader.getExtension extension_spec: |%s|" % extension_spec
        subdir = 'extension'
        module_name = None
        # split the extension_spec on '.'
        if '.' not in extension_spec:
            extension_class_name = extension_spec
        elif extension_spec.count('.') == 1:
            module_name, extension_class_name = extension_spec.split('.')
        elif extension_spec.count('.') == 2:
            subdir, module_name, extension_class_name = extension_spec.split('.')
        else:
            raise Exception('Invalid extension specification: %s' % extension_spec)

        return super(ExtensionLoader, self).getFacility('Extension', extension_class_name,
                                                        subdir=subdir,
                                                        module_name=module_name)

