# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import os, sys
import inspect
import traceback

######################################################################################

errout = sys.stderr.write

_blurt = False
_logger = None

######################################################################################

def logAllExceptions(truthiness, logger=None):
    """
        Call this function when you want the convenience of raising the
        Exception subclasses defined in this module without having to supply
        an instance of a logger for each raise statement.
        Simply supply the logger to be used in this call and you're set
        for the duration!
    """
    global _blurt, _logger
    if truthiness and logger:
        _blurt = True
        _logger = logger
    else:
        _blurt = False
        _logger = None

######################################################################################

def getCaller(frame):
    """
        Given a stack frame object, return the name of the function
        or the name of the class and method that is operating in the frame.
    """
    caller = ""
    function_name = frame.f_code.co_name

    # Get the name of the calling class
    if "self" in frame.f_locals:
        caller_frame_object = frame.f_locals["self"]
        obj_name = None
        if hasattr(caller_frame_object, "__instance_name__"):
            obj_name = caller_frame_object.__instance__name__()
        elif hasattr(caller_frame_object, "__class__"):
            obj_name = caller_frame_object.__class__.__name__
        else:
            obj_name = caller_frame_object.__name__

        if obj_name:
            caller = "%s%s." % (caller, obj_name)

    # caller + function_name

    # Give a name to the function that calls all
    caller = caller.replace("?", "main")

    return caller + function_name

def getTracebackInfo():
    tb = sys.exc_info()[2]
    tb_info = traceback.format_tb(tb)[0].split(',')
    problem_file = os.path.basename(tb_info[0].split('File ')[1].strip('"'))
    prob_line_no = tb_info[1]
    meth_name    = tb_info[2].split("\n", 1)[0].replace('in ','') + '()'
    fault_ident = "problem occurred in %s %s:%s" % (problem_file, meth_name, prob_line_no)
    return fault_ident

def stackulus(stack_item):
    """
        placeholder for code than can futz with various levels of the stack and
        retrieve interesting things about an item at particular stack level
    """
    frame, filename, line_no, code_unit = stack_item[:4]
    caller = getCaller(frame)
    return caller, filename, line_no, code_unit

######################################################################################

class EIFException(Exception):
    def __init__(self, msg, logger=None):
        self.msg     = msg
        self.logger = logger

        if hasattr(self,'ex_type') and self.ex_type == 'Configuration':
            # use inspect, or other module to look at my caller (grab file, class, line) so that we can
            # set some instance var to be used when report is called (instead of the std report
            # handling which uses the result of getTracebackInfo()
            self._recordFingerprint()
        raise self

    def report(self):
        fault = getTracebackInfo()
        #print(fault)
        global _blurt, _logger
        if self.logger or (_blurt and _logger):
            log = self.logger or _logger
            log_method = getattr(log, self.log_level, 'error')
            # prepend message with specific exception class name
            msg = "<%s> %s" % (self.__class__.__name__, self.msg)
            log_method(fault, exception_triggered=True)
            log_method(msg, exception_triggered=True)
        else:
            errout('%s raised: %s\n%s' % (self.__class__.__name__, self.msg, fault))
            sys.stderr.flush()
        #
        # should we blurt out any traceback info?
        #
        # stack_level could be 2 or 3 or 4 or 5 ...
        # stack_level = 5  # this seems the right value in py.test mode
        # stack_level = 3  # this seems the right value in regular execution mode
        # caller, filename, line_no, code_unit = stackulus(inspect.stack()[stack_level])
        # caller += "(%d)" % line_no


class FatalError(EIFException):
    """
        This brand of EIFException is intended be used when an operation has failed and
        no further processing can or should be attempted.
        The condition that caused this operation will typically be caused by an
        initialization/setup error where some part of the overall machinery is missing
        or unable to be exercised.
    """
    log_level = 'fatal'


class ConfigurationError(EIFException):
    """
        This brand of EIFException is intended to be used when a configuration error
        has been detected. Because of the configuration error no further processing
        can or should proceed. The configuration will have to be fixed or this
        exception would be raised again in a future invocation.
    """
    log_level = 'fatal'

    def __init__(self, msg, logger=None):
        self.ex_type = 'Configuration'
        self.cause_file = None
        self.cause_class = None
        self.cause_method = None
        self.cause_line = None
        super().__init__(msg, logger=None)

    def _recordFingerprint(self):
        caller, filename, line_no, code_unit = stackulus(inspect.stack()[3])
        self.cause_file, self.cause_method, self.cause_line = filename, caller, line_no

    def report(self):
        fault_location = "%s %s(%d)" % (self.cause_method, os.path.basename(self.cause_file), self.cause_line)
        log_message = "%s - %s" % (fault_location, self.msg)
        global _blurt, _logger
        if self.logger or (_blurt and _logger):
            log = self.logger or _logger
            log_method = getattr(log, self.log_level, 'fatal')
            #msg = "<%s> %s" % (self.__class__.__name__, log_message)
            log_method(log_message, exception_triggered=True)
        else:
            errout('%s raised: %s\n%s' % (self.__class__.__name__, self.msg, log_message))
            sys.stderr.flush()


class NonFatalConfigurationError(EIFException):
    """
        This brand of EIFException is intended to be used when a configuration error
        has been detected. This exception differs from the ConfigurationError
        in that the configuration error detected will not prevent reasonable operation
        from proceeding but should be noted for future correction.
    """
    log_level = 'error'


class OperationalError(EIFException):
    """
        This brand of EIFException is intendedto be used when an operation has failed and
        no further processing can or should be attempted.
        The condition that caused this though may not be in place if the operation were
        attempted again at a later time.
    """
    log_level = 'error'


__all__ = [  ConfigurationError, NonFatalConfigurationError, OperationalError,  logAllExceptions ]
