# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

# This is the "large stone" level operation of the ghconn infrastructure,
# wherein the two connections held by the connector get instantiated, 
# connected to their respective systems and exercised.

import sys, os, platform
import re
from collections import OrderedDict

from vcsconn.utils.eif_exception   import FatalError, ConfigurationError, OperationalError
from vcsconn.utils.claslo          import ClassLoader

##############################################################################################

__version__ = "2.0.5"

PLUGIN_SPEC_PATTERN       = re.compile(r'^(?P<plugin_class>\w+)\s*\((?P<plugin_config>[^\)]*)\)\s*$')
PLAIN_PLUGIN_SPEC_PATTERN = re.compile(r'(?P<plugin_class>\w+)\s*$')

ARCHITECTURE_PKG = 'vcsconn'

VALID_AUTHOR_TRANSFORMS = ['Passthrough','FileBasedUserNameLookup','UserNameDomainAugmentLookup']

##############################################################################################

class VCSConnector:

    def __init__(self, config, logger):
        self.config = config
        self.log    = logger

        conn_sections = []
        try:
            conn_sections = [section for section in config.topLevels()
                                      if section not in ['Service']]
            if len(conn_sections) != 2:
                raise ConfigurationError('Config does not contain two connection sections')
        except Exception:
            raise ConfigurationError('Config file lacks sufficient information for VCSConnector')

        self.agicen_conn = None
        self.vcs_conn    = None

        self.vcs_name = [name for name in conn_sections if not name.startswith('AgileCentral')][0]
        self.log.info("Agile Central VCS Connector for %s, version %s" % (self.vcs_name, __version__))
        uname_fields = platform.uname()
        self.log.info("Python platform: %s %s" % (uname_fields[0], uname_fields[2]))
        self.log.info("Python  version: %s" % sys.version.split()[0])
        self.transforms = None

        self.internalizeConfig(config)

        droid = ClassLoader()
        agicen_conn_class_name = config.connectionClassName('AgileCentral')
        vcs_conn_class_name    = config.connectionClassName(self.vcs_name)

        try:
            self.agicen_conn_class = droid.loadConnectionClass(agicen_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load AgileCentralVCSConnection class, %s' % msg)
        try:
            self.vcs_conn_class = droid.loadConnectionClass(vcs_conn_class_name, pkgdir=ARCHITECTURE_PKG)
        except Exception as msg:
            raise FatalError('Unable to load %sConnection class, %s' % (self.vcs_name, msg))

        self.establishConnections()

        if not self.validate():  # basically just calls validate on both connection instances
            raise ConfigurationError("Validation failed")
        self.log.info("Initialization complete: Delegate connections operational, ready for scan/reflect ops")

    def internalizeConfig(self, config):
        self.agicen_conf = config.topLevel('AgileCentral')
        self.vcs_conf    = config.topLevel(self.vcs_name)
        self.svc_conf    = config.topLevel('Service')
        #self.max_commits = self.svc_conf.get('MaxCommits', 20)
        self.lookback    = self.svc_conf.get('Lookback', 3601)
        valid_config_items = ['Preview', 'LogLevel', 'SecurityLevel', 'Lookback', 'PullRequests', 'AppendVCSCommitter']
        svc_conf = config.topLevel('Service')
        invalid_config_items = [item for item in svc_conf.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "Service section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)

        # if "Repository" in self.vcs_conf:
        #     # look for include and exclude
        #     print("found Repository in vcs_conf")
        #     pass


    def establishConnections(self):
        self.agicen_conn = self.agicen_conn_class(self.agicen_conf, self.log)
        self.vcs_conn    = self.vcs_conn_class(self.vcs_conf, self.log)
        self.vcs_conn.connect()  # we do this before agicen_conn to be able to get the GitHub backend version

        if self.agicen_conn and self.vcs_conn and getattr(self.agicen_conn, 'set_integration_header'):
            agicen_headers = {'name'    : 'Agile Central VCSConnector for %s' % self.vcs_name,
                              'version' : __version__,
                              'vendor'  : 'CA Technologies',
                              'other_version' : self.vcs_conn.backend_version
                            }
            self.agicen_conn.set_integration_header(agicen_headers)
        self.agicen_conn.setSourceIdentification(self.vcs_conn.name(), self.vcs_conn.backend_version)
        self.agicen_conn.connect()  # so we can use it in our X-Rally-Integrations header items here


    def validate(self):
        """
            This calls the validate method on both the Agile Central and the VCS connections
        """
        self.log.info("Connector validation starting")

        if not self.agicen_conn.validate():
            self.log.info("AgileCentralConnection validation failed")
            return False
        self.log.info("AgileCentralConnection validation succeeded")

        if not self.vcs_conn.validate():
            self.log.info("%sConnection validation failed" % self.vcs_name)
            return False
        self.log.info("%sConnection validation succeeded" % self.vcs_name)


        self.log.info("Connector validation completed")

        return True


    def run(self, last_run, extension):
        """
            The real beef is in the call to reflectChangesetsInAgileCentral.
            The facility for extensions is not yet implemented for VCS connectors,
            so the pre and post batch calls are currently no-ops.
        """
        self.preBatch(extension)
        cs_status = True
        pr_status = True
        pull_requests = None
        cs_status, commits = self.reflectChangesetsInAgileCentral(last_run)
        self.postBatch(extension, cs_status, commits)
        if self.svc_conf.get('PullRequests', False):
            pr_status, pull_requests = self.reflectPullRequestsInAgileCentral(last_run)
        overall_status = cs_status and pr_status

        return overall_status, commits, pull_requests


    def preBatch(self, extension):
        """
        """
        if extension and 'PreBatchAction' in extension:
            preba = extension['PreBatchAction']
            preba.service()


    def postBatch(self, extension, status, commits):
        """
        """
        if extension and 'PostBatchAction' in extension:
            postba = extension['PostBatchAction']
            postba.service(status, commits)

    def reflectChangesetsInAgileCentral(self, ref_time):
        """
            The last run time is passed to Connection objects in UTC; it is an epoch seconds value.
            they are responsible for converting if necessary.
            Time in log messages is always reported in UTC (aka Z or Zulu time).
        """
        status = True
        agicen = self.agicen_conn
        vcs = self.vcs_conn

        preview_mode         = self.svc_conf.get('Preview', False)
        append_vcs_committer = self.svc_conf.get('AppendVCSCommitter', False)

        if preview_mode:
            pm_tag = "Preview: "
            self.log.info('***** Preview Mode *****   (no Changesets will be created in Agile Central)')
            return status, []

        projects = []
        projects.append(self.agicen_conf.get('Project',None))
        recent_agicen_commits = agicen.getRecentCommits(ref_time.minus(self.lookback, units='seconds'))
        recent_vcs_commits    =    vcs.getRecentCommits(ref_time.minus(self.lookback, units='seconds'))
        recorded_commits = OrderedDict()

        if recent_vcs_commits:
            unrecorded_commits = self._identifyUnrecordedChangesets(recent_agicen_commits, recent_vcs_commits)
            uncomms_count = sum([len(changesets) for changesets in unrecorded_commits.values()])
            self.log.info("total unrecorded Changesets count: %d (in %d repositories)" % (uncomms_count, len(unrecorded_commits)))
            try:
                if unrecorded_commits:
                    recorded_commits = agicen.postCommits(unrecorded_commits, append_vcs_committer)
                    status = True
            except Exception as msg:
                self.log.error('OperationalError postCommits - %s' % msg)
                status = False

        return status, recorded_commits


    def _showChangesetInformation(self, agicen_commits, vcs_commits):
        ##
        for project, job_commits in agicen_commits.items():
            print("Agile Central project: %s" % project)
            for job, commits in job_commits.items():
                print("    %-36.36s : %3d build items" % (job, len(commits)))
        print("")

        ##
        for view, job_commits in vcs_commits.items():
            print("Jenkins View: %s" % view)
            for job, commits in job_commits.items():
                print("    %-36.36s : %3d build items" % (job, len(commits)))
        print("")
        ##


    def _identifyUnrecordedChangesets(self, agicen_commits, vcs_commits):
        """
            If there are items in the agicen_commits for which there is  a counterpart in
            the vcs_commits, the information has already been reflected in Agile Central.  --> NOOP

            If there are items in the vcs_commits   for which there is no counterpart in
            the agicen_commits, those items are candidates to be reflected in Agile Central --> REFLECT

            If there are items in the agicen_commits for which there is no counterpart in
            the vcs_commits, information has been lost,  dat would be some bad... --> ERROR
        """
        unrecorded_commits = {}
        vcs_commit_shas    = {}
        agicen_commit_shas = {}
        repo_names = list(vcs_commits.keys())
        for repo_name in repo_names:
            vcs_commit_shas[repo_name]     = [commit.rev_num for commit in vcs_commits[repo_name]]
            agicen_commit_shas[repo_name]  = [commit.Revision for commit in agicen_commits if commit.SCMRepository.Name == repo_name]
            diff = list(set(vcs_commit_shas[repo_name]) - set(agicen_commit_shas[repo_name]))
            if diff:
                unrecorded_commits[repo_name] = [commit for commit in vcs_commits[repo_name] if commit.rev_num in diff]
        return unrecorded_commits

    def _identifyUnrecordedPullRequests(self, agicen_prs, vcs_prs):
        """
            If there are items in the agicen_prs for which there is  a counterpart in
            the vcs_prs, the information has already been reflected in Agile Central.  --> NOOP

            If there are items in the vcs_prs   for which there is no counterpart in
            the agicen_prs, those items are candidates to be reflected in Agile Central --> REFLECT

            If there are items in the agicen_prs for which there is no counterpart in
            the vcs_prs, information has been lost,  dat would be some bad... --> ERROR
        """
        agicen_pr_idents = [ac_pr.ExternalID for ac_pr in agicen_prs]
        unrecorded_prs   = [vcs_pr for vcs_pr in vcs_prs if vcs_pr.ident not in agicen_pr_idents]
        return unrecorded_prs

    def reflectPullRequestsInAgileCentral(self, ref_time):
        """
                   The last run time is passed to Connection objects in UTC; it is an epoch seconds value.
                   they are responsible for converting if necessary.
                   Time in log messages is always reported in UTC (aka Z or Zulu time).
               """
        status = True
        agicen = self.agicen_conn
        vcs = self.vcs_conn

        preview_mode = self.svc_conf.get('Preview', False)

        if preview_mode:
            pm_tag = "Preview: "
            self.log.info('***** Preview Mode *****   (no PullRequests will be created in Agile Central)')
            return status, []

        projects = []
        projects.append(self.agicen_conf.get('Project', None))
        recent_agicen_prs = agicen.getRecentPullRequests(ref_time.minus(self.lookback, units='seconds'))
        recent_vcs_prs    =    vcs.getRecentPullRequests(ref_time.minus(self.lookback, units='seconds'))
        recorded_pull_requests = OrderedDict()

        if recent_vcs_prs:
            unrecorded_pull_requests = self._identifyUnrecordedPullRequests(recent_agicen_prs, recent_vcs_prs)
            self.log.info("unrecorded PullRequests count: %d" % len(unrecorded_pull_requests))
            try:
                if unrecorded_pull_requests:
                    recorded_pull_requests = agicen.postPullRequests(unrecorded_pull_requests)
                    status = True
            except Exception as ex:
                self.log.error('OperationalError postPullRequests - %s' % ex)
                status = False

        return status, recorded_pull_requests


####################################################################################
