# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

############################################################################################################
#
#  vcs_connector_runner - control machinery that detects commits in a version control system (VCS) and
#                         reflects those Changesets in a AC Agile Central subscription.
#
############################################################################################################

import sys, os
import re
import time
import glob
from calendar import timegm

from vcsconn.utils.klog       import ActivityLogger
from vcsconn.utils.proctbl    import ProcTable
from vcsconn.utils.lock_file  import LockFile
from vcsconn.utils.time_file  import TimeFile
from vcsconn.utils.konfabulus import Konfabulator
from vcsconn.vcs_connector    import VCSConnector
from vcsconn.utils.eif_exception import ConfigurationError, NonFatalConfigurationError, FatalError
from vcsconn.utils.eif_exception import OperationalError, logAllExceptions
from vcsconn.utils.chronuti import TimeStamp, ISO_FORMAT, TIMEFILE_FORMAT

############################################################################################################

ARCHITECTURE_ACRONYM = 'eif'
__version__ = "2.0.5"

EXISTENCE_PROCLAMATION = """
************************************************************************************************************************

     CA Agile Central VCSConnector for %s  version %s  starting at: %s  with pid: %s
           curdir: %s
          command: %s

**********************************************************************************f**************************************
"""

LOCK_FILE  = 'LOCK.tmp'
STD_TS_FMT = '%Y-%m-%d %H:%M:%S Z'

############################################################################################################

class VCSConnectorRunner:

    def __init__(self, args):
        """
            An instance of this class is responsible for basic command line parsing,
            pulling off the name of the target config file.
            Once that item is taken care of, the instance obtains an instance of a
            Konfabulator, a AuxiliaryClassLoader instance and and instance of an VCSConnector.
            The Konfabulator is delegated the task of reading the config file, the AuxiliaryClassLoader
            is delegated the task of pulling in any Extension classes.
            These instances are then provided to the VCSConnector instance.
            This instance then directs the VCSConnector instance to obtain unrecorded commits
            from the target VCS system and reflect them in the Agile Central server.
        """

        options = [item for item in args if item.startswith('-')]
        args = [item for item in args if item not in options]

        if len(args) < 1:
            problem = ("Insufficient command line args, must be at least a config file name.")
            raise ConfigurationError(problem)

        self.default_log_file_name = True
        self.cleartext_flag = False
        # this only allows a logfile spec of --logfile=<some_file_name>
        log_spec = [opt for opt in options if opt.startswith('--logfile=')]
        if log_spec:
            self.logfile_name = log_spec[0].replace('--logfile=', '')
            self.default_log_file_name = False
            for spec in log_spec:
                args = [arg for arg in args if arg != spec]


        clear_text_creds = [opt for opt in options if opt in ['-c', '--c'] or opt.startswith('--cleartext')]
        if clear_text_creds:
            self.cleartext_flag = True
        self.config_file_names = args
        if self.default_log_file_name:  # set it to first config file minus any '_config.yml' portion
            self.first_config = self.config_file_names[0]
            self.logfile_name = "log/%s.log" % self.first_config.replace('.yml', '').replace('_config', '')
            try:
                if not os.path.exists('log'):
                    os.makedirs('log')
            except Exception as msg:
                sys.stderr.write("Unable to locate or create the log sub-directory, %s\n" % msg)
                raise Exception

        self.log = ActivityLogger(self.logfile_name)
        logAllExceptions(True, self.log)
        self.preview   = False
        self.connector = None
        self.extension = {}


    def proclaim_existence(self, vcs_system_name):
        proc = ProcTable.targetProcess(os.getpid())
        now = time.strftime(STD_TS_FMT, time.gmtime(time.time()))
        cmd_elements = proc.cmdline.split()
        executable = cmd_elements[0]
        cmd_elements[0] = os.path.basename(executable)
        proc.cmdline = " ".join(cmd_elements)
        self.log.write(EXISTENCE_PROCLAMATION % (vcs_system_name, __version__, now, proc.pid, os.getcwd(), proc.cmdline))

    def acquireLock(self):
        """
            Check for conditions to proceed based on lock file absence/presence status.
            Acquisition of the lock is a green-light to proceed.
            Failure to acquire the lock prevents any further operation.
        """
        if LockFile.exists(LOCK_FILE):
            self.log.warning("A %s file exists" % LOCK_FILE)
            locker = LockFile.currentLockHolder(LOCK_FILE)
            if not LockFile.lockerIsRunning(LOCK_FILE, locker):
                message = ("A prior connector process (%s) did not clean up "
                           "the lock file on termination, proceeding with this run")
                self.log.warning(message % locker)
            else:
                problems = ["Another connector process [%s] is still running, unable to proceed" % locker,
                            "Simultaneous processes for this connector are prohibited" ]
                raise FatalError(problems)

        try:
            LockFile.createLock(LOCK_FILE)
        except:
            problems = ["Unable to create a process lock file: %s, check permissions in the connector root directory" % LOCK_FILE ]
            raise FatalError(problems)
        return True

    def releaseLock(self):
        LockFile.destroyLock(LOCK_FILE)

    def run(self):
        self.acquireLock()

        for config_file in self.config_file_names:
            try:
                config_file_path = self.find_config_file(config_file)
                if not config_file_path:
                    self.log.error("No config file for '%s' found in the config subdir" % config_file)
                    continue
                lf_name = "log/%s.log" % config_file.replace('.yml', '').replace('_config', '')
                self.log = ActivityLogger(lf_name)
                logAllExceptions(True, self.log)
                self._operateService(config_file_path)
                success = True
            except ConfigurationError as ex:
                ex.report()
                success = False
            except Exception as ex:
                self.log.error(ex)
                success = False

            if success:
                self.log.info('run completed for %s' % config_file)

        try:
            self.releaseLock()
        except Exception as msg:
            raise OperationalError("ERROR: unable to remove lock file '%s', %s" % (LOCK_FILE, msg))


    def identifyVCSName(self, config):
        return config.topLevelTargetSystemName()

    def find_config_file(self, config_name):
        relative_path = 'config/%s' % config_name
        if os.path.exists(relative_path):
            return relative_path
        else:
            return None

    def _operateService(self, config_file_path):
        config_name = config_file_path.replace('config/', '')
        self.connector = None

        config = self.getConfiguration(config_name)
        if not config:
            return
        target_vcs_name = self.identifyVCSName(config)
        self.proclaim_existence(target_vcs_name)
        self.log.info("processing to commence using content from %s" % config_file_path)
        last_conf_mod = time.strftime(STD_TS_FMT, time.gmtime(os.path.getmtime(config_file_path)))
        conf_file_size = os.path.getsize(config_file_path)
        self.log.info("%s last modified %s,  size: %d chars" % (config_file_path, last_conf_mod, conf_file_size))

        if self.cleartext_flag:
            self.log.info("Running connector with cleartext flag. SecurityLevel will not be applied to credentials in cleartext")
        config.applySecurityPolicy()

        this_run = TimeStamp.now()

        self.time_file = TimeFile(self.vcsTimeFileName(config_name), self.log)
        if self.time_file.exists():
            last_run = self.time_file.read() # the time_file read method returns an epoch seconds value
        else:
            last_run = this_run.minus(3, units='days')
        self.log.info("Time File value %s --- Now %s" % (last_run.asLogStamp(), this_run.asLogStamp()))

        self.connector = VCSConnector(config, self.log)
        self.log.debug("Got a VCSConnector instance, calling the VCSConnector.run ...")
        status, changesets, pull_requests = self.connector.run(last_run, self.extension)
        # changesets is an OrderedDict instance, keyed by repository name, value is a list of Changeset instances

        elapsed = TimeStamp.now().minus(this_run.asEpochSeconds(), units='seconds')
        self.logServiceStatistics(config_name, changesets, elapsed)

        if self.preview:
            self.log.info("Preview mode in effect, time.file File not written/updated")
            return
        if not status:
            # Not writing the time.file may cause repetitive detection of Changesets,
            # but that is better than missing out on Changesets altogether
            self.log.info("There was an error in processing so the time.file was not written")
            return

        if not changesets and not pull_requests:
            self.log.info('No Changesets or PullRequests were added during this run, so the time.file NOT updated.')
            return

        last_run_ts = self.writeTimeFile(changesets, pull_requests)
        return

    def writeTimeFile(self, repo_changesets, pull_requests):
        try:
            latest_values = []
            if repo_changesets:
                for repo, changesets in repo_changesets.items():
                    latest_values.append(max([cs.CommitTimestamp for cs in changesets]))
            if pull_requests:
                latest_values.append(max([pr.CreationDate for pr in pull_requests]))
            candidate_timestamp = min(latest_values)

            time_file_timestamp = candidate_timestamp.replace('T', ' ')[:19]
            self.time_file.write(time_file_timestamp)
            self.log.info("time file written with value of %s Z" % time_file_timestamp)
            return time_file_timestamp
        except Exception as msg:
            raise OperationalError(msg)

    def vcsTimeFileName(self, config_file):
        if config_file:
            if config_file.endswith('.yml') or  config_file.endswith('.cfg'):
                time_file_name = config_file[0:-4] + '_time.file'
            else:
                time_file_name =  "%s_time.file" % config_file
        else:
            time_file_name = 'time.file'
        time_file_path = 'log/%s' % time_file_name
        return time_file_path


    def logServiceStatistics(self, config_name, changesets, elapsed):
        """
            what we intend to append to the log...  
               vcs_conn_config.yml: 32 additional Changesets reflected in AgileCentral
            and a line with the elapsed time taken in human readable form  
            (elapsed is in seconds (a float value))
        """
        for changeset in changesets:
            preview_reminder = "(Preview Mode)" if self.preview else ""
            factoid = "%3d Changesets posted for repository %s" % (len(changesets[changeset]), changeset)
            self.log.info("%s: %s %s" % (config_name, factoid, preview_reminder))
        hours, rem = divmod(elapsed.asEpochSeconds(), 3600)
        mins, secs = divmod(rem, 60)
        if hours > 0:
            duration = "%s hours %s minutes %4.2 seconds" % (hours, mins, float(secs))
        else:
            if mins > 0:
                duration = "%s minutes %4.2f seconds" % (mins, float(secs))
            else:
                duration = "%4.2f seconds" % float(secs)
        self.log.info("%s: service run took %s" % (config_name, duration))


    def getConfiguration(self, config_file):
        config = None
        try:
            config = Konfabulator(config_file, self.log, self.cleartext_flag)
        except NonFatalConfigurationError as ex:
            ex.report()
            return None # info for this will have already been logged or blurted
        except Exception as msg:
            raise ConfigurationError(msg)
        svc_conf = config.topLevel('Service')
        self.preview = False
        if svc_conf and svc_conf.get('Preview', None) == True:
            self.preview = True
        self.log_level = 'Info'
        if svc_conf:
            ll = svc_conf.get('LogLevel', 'Info').title()
            if ll in ['Fatal', 'Error', 'Warn', 'Info', 'Debug']:
                self.log_level = ll
                self.log.setLevel(self.log_level)
            else:
                pass # bad LogLevel specified
            #if 'PostBatchExtension' in svc_conf:
            #    pba_class_name = svc_conf['PostBatchExtension']
            #    pba_class = ExtensionLoader().getExtension(pba_class_name)
            #    self.extension['PostBatch'] = pba_class()

        return config

